"""Apply data reduction to HDF5 files structured in the European XFEL Data Format (EXDF).
"""

from argparse import ArgumentParser
from collections import defaultdict
from pathlib import Path
import fnmatch
import logging
import re
import sys

from pkg_resources import iter_entry_points
from extra_data import RunDirectory, open_run, by_id
from extra_data.writer import FileWriter, VirtualFileWriter
from extra_data.read_machinery import select_train_ids


exdf_filename_pattern = re.compile(
    r'([A-Z]+)-R(\d{4})-(\w+)-S(\d{5}).h5')

def get_source_origin(sourcedata):
    # TODO: Can be replaced by EXtra-data functionality later.

    try:
        files = sourcedata.source_files
    except AttributeError:
        files = sourcedata.files

    storage_classes = set()
    run_numbers = set()
    aggregators = set()

    for f in files:
        m = exdf_filename_pattern.match(Path(f.filename).name)

        if m:
            storage_classes.add(m[1])
            run_numbers.add(int(m[2]))
            aggregators.add(m[3])

    result_sets = [storage_classes, run_numbers, aggregators]

    if any([len(x) != 1 for x in result_sets]):
        raise ValueError('different or no classes or aggregators recognized '
                         'from source filenames')

    return storage_classes.pop(), run_numbers.pop(), aggregators.pop()


class ReductionSequence:
    # A sequence of reduction methods.

    log = logging.getLogger('reduce.ReductionSequence')

    def __init__(self, data, methods):
        self._data = data
        self._methods = methods

        self._instructions = sum(
            [m._instructions for m in methods.values()], [])

        self._sources = sorted(data.all_sources)
        self._touched_sources = set()

        # Only populated if trains/keys are removed for select sources.
        self._custom_keys = {}  # source -> set(<keys>)
        self._custom_trains = {}  # source -> list(<trains>)
        self._partial_copies = {}  # (source, key) -> (region, chunks)

    def _filter_ops(self, op):
        return [args[1:] for args in self._instructions if args[0] == op]

    def link(self, output_scope):
        """Link together multiple methods into a sequence."""

        for source_glob, in self._filter_ops('removeSources'):
            for source in fnmatch.filter(self._sources, source_glob):
                self._touched_sources.add(source)
                self._sources.remove(source)                

        for source_glob, key_glob in self._filter_ops('removeKeys'):
            for source in fnmatch.filter(self._sources, source_glob):
                self._touched_sources.add(source)

                keys = self._custom_keys.setdefault(
                    source, set(self._data[source].keys()))

                for key in fnmatch.filter(keys, key_glob):
                    keys.remove(key)

        for source_glob, train_sel in self._filter_ops('selectTrains'):
            for source in fnmatch.filter(self._sources, source_glob):
                self._touched_sources.add(source)
                train_ids = self._custom_trains.setdefault(
                    source, list(self._data.train_ids))

                self._custom_trains[source] = select_train_ids(
                    self._custom_trains[source], train_sel)

        for source_glob, key, region, chunking in self._filter_ops('partialCopy'):
            for source in fnmatch.filter(self._sources, source_glob):
                self._touched_sources.add(source)
                self._partial_copies[(source, key)] = (region, chunking)

        if output_scope == 'sources':
            self._sources = sorted(
                self._touched_sources.intersection(self._sources))

        elif output_scope == 'aggregators':
            touched_aggregators = {get_source_origin(self._data[source])[2]
                                   for source in self._touched_sources}

            self._sources = sorted(
                {source for source in self._sources
                if get_source_origin(self._data[source])[2]
                    in touched_aggregators})

        if not self._sources:
            self.log.warn('Linking reduction methods yielded empty source '
                          'selection')

    def write_collection(self, output_path, sequence_len=512):
        outp_data = self._data.select([(s, '*') for s in self._sources])
        origin_sources = defaultdict(list)

        print(len(self._data.train_ids), len(outp_data.train_ids))
        return

        run_numbers = set()

        # Collect all origins (combination of storage class and aggregator)
        # and the sources they contain.
        for source in self._sources:
            storage_class, run_no, aggregator = get_source_origin(
                outp_data[source])
            origin_sources[(storage_class, aggregator)].append(source)
            run_numbers.add(run_no)

        if len(run_numbers) != 1:
            self.log.error('Could not infer run number from filenames')
            return

        run_no = run_numbers.pop()

        for (storage_class, aggregator), sources in origin_sources.items():
            origin_tids = sorted(set(outp_data.train_ids).intersection(*[
                tids for s, tids in self._custom_trains.items()
                if s in sources]))
            origin_sel = {
                s: self._custom_keys[s] if s in self._custom_keys else set()
                for s in sources}

            # Select output data down to what's in this origin.
            origin_data = outp_data \
                .select_trains(by_id[origin_tids]) \
                .select(origin_sel)

            num_trains = len(origin_data.train_ids)
            num_sequences = int(num_trains / sequence_len) + 1

            self.log.info(
                f'{storage_class}-{aggregator}: '
                f'{len(origin_data.all_sources)} sources with {num_trains} '
                f'trains over {num_sequences} sequences')

            for seq_no in range(num_sequences):
                train_slice = slice(
                    seq_no * sequence_len,
                    (seq_no + 1) * sequence_len)

                # Select origin data down to what's in this file.
                file_data = origin_data.select_trains(train_slice)

                # file_data now can contain trains for which no source has
                # any data. .select(require_all) Unfortunately does the
                # opposite of what we want

                file_path = output_path / '{}-R{:04d}-{}-S{:06d}.h5'.format(
                    storage_class, run_no, aggregator, seq_no)

                self.log.debug(
                    f'{file_path.stem}: {len(file_data.all_sources)} sources '
                    f'with {len(file_data.train_ids)} trains')
                fw = ReductionFileWriter(file_path, file_data, self)
                fw.write()

    def write_voview(self, output_path):
        raise NotImplementedError('voview output layout')

    def write_collapsed(self, output_path, sequence_len):
        raise NotImplementedError('collapsed collection output layout')


class ReductionMethod:
    log = logging.getLogger('ReductionMethod')

    # Stores list of instructions and can apply them
    # Allows (de-)serialization with files
    # Convenience interface for reduction methods to emit instructions.

    # Populate some arguments
    # Check if the method is enabled -> could be in the main body
    # Actually compile instructions for removal with DataCollection

    def __init__(self):
        self._instructions = []

    def __bool__(self):
        # Whether this program contains any instructions.
        return bool(self._instructions)

    def __len__(self):
        # How many instructions this program contains.
        return len(self._instructions)

    def select_trains(self, sources, trains):
        # init level
        self._instructions.append(('selectTrains', sources, trains))
        self.log.debug(f'Emitted {self._instructions[-1]}')

    def remove_sources(self, source_glob):
        # init level
        self._instructions.append(('removeSources', source_glob))
        self.log.debug(f'Emitted {self._instructions[-1]}')

    def remove_keys(self, source_glob, key_glob):
        # copy_dataset level
        self._instructions.append(('removeKeys', source_glob, key_glob))
        self.log.debug(f'Emitted {self._instructions[-1]}')

    def partial_copy(self, sources, key, region, chunks):
        self._instructions.append(
            ('partialCopy', sources, key, region, chunks))
        self.log.debug(f'Emitted {self._instructions[-1]}')


class ReductionFileWriter(FileWriter):
    def __init__(self, path, data, sequence):
        super().__init__(path, data)
        self.sequence = sequence

    def prepare_source(self, source):
        for key in sorted(self.data.keys_for_source(source)):
            path = f"{self._section(source)}/{source}/{key.replace('.', '/')}"
            nentries = self._guess_number_of_storing_entries(source, key)
            src_ds1 = self.data[source].files[0].file[path]

            try:
                chunking_kwargs = dict(
                    chunks=self.sequence._partial_copies[source, key][1])
            except KeyError:
                chunking_kwargs = dict()

            self.file.create_dataset_like(
                path, src_ds1, shape=(nentries,) + src_ds1.shape[1:],
                # Corrected detector data has maxshape==shape, but if any max
                # dim is smaller than the chunk size, h5py complains. Making
                # the first dimension unlimited avoids this.
                maxshape=(None,) + src_ds1.shape[1:],
                **chunking_kwargs
            )
            if source in self.data.instrument_sources:
                self.data_sources.add(f"INSTRUMENT/{source}/{key.partition('.')[0]}")

        if source not in self.data.instrument_sources:
            self.data_sources.add(f"CONTROL/{source}")        

    def copy_dataset(self, source, key):
        path = f"{self._section(source)}/{source}/{key.replace('.', '/')}"

        try:
            region = self.sequence._partial_copies[source, key][0]
        except KeyError:
            a = self.data.get_array(source, key)
            self.file[path][:] = a.values
        else:
            a = self.data.get_array(source, key)
            full_region = (slice(None, None, None), *region)
            
            self.file[path][full_region] = a.values[full_region]
        
        self._make_index(source, key, a.coords['trainId'].values)


proposal_run_input_pattern = re.compile(
    r'^p(\d{4}|9\d{5}):r(\d{1,3}):*([a-z]*)$')


def main(argv=None):
    ap = ArgumentParser(
        description='Apply data reduction to HDF5 files structured in the '
                    'European XFEL Data Format (EXDF).'
    )

    ap.add_argument(
        'input', metavar='INPUT', type=Path,
        help='folder for input data or pX:rY[:prefix=raw] to load a proposal '
             'run from /gpfs/exfel/exp directly, currently always assumed to '
             'be a single run.'
    )

    logging_group = ap.add_mutually_exclusive_group()

    logging_group.add_argument(
        '--verbose', '-v', action='store_true',
        help='set logging filter to DEBUG'
    )

    logging_group.add_argument(
        '--quiet', '-q', action='store_true',
        help='set logging filter to WARNING'
    )

    logging_group.add_argument(
        '--silent', '-s', action='store_true',
        help='set logging filter to ERROR'
    )

    for ep in iter_entry_points('exdf_tools.reduction_method'):
        ep.load().arguments(ap)

    output_group = ap.add_argument_group(
        'output arguments',
        'Allows to configure the output location, scope and layout.'
    )

    output_group.add_argument(
        '--output', '-o', metavar='PATH', type=Path, action='store',
        default='./out',
        help='folder for output data, or file if ending with .h5 '
              'and used with \'voview\' output layout')

    output_group.add_argument(
        '--output-scope', action='store',
        choices=['sources', 'aggregators', 'all', 'none'],
        default='aggregators',
        help='scope of what is contained in output, either only the '
             '\'sources\' touched by data reduction, the \'aggregator\' '
             'touched, \'all\' data found in the input or \'none\' of it '
             'resulting in no actual output data files'
    )

    output_group.add_argument(
        '--output-layout', action='store',
        choices=['collection', 'voview', 'collapsed'],
        default='collection',
        help='layout of output files, either as a \'collection\' of separate '
             'aggregator and sequence files, as a \'voview\' file containing '
             'virtual datasets for all data or a single \'collapsed\' '
             'collection combining all former aggregators into one'
    )

    output_group.add_argument(
        '--output-sequence-len', action='store', type=int, default=256,
        help='number of trains per sequence file or -1 to preserve input '
             'length, only used for \'collection\' and \'collapsed\' output '
             'layout'
    )

    output_group.add_argument(
        '--to-script', action='store', type=Path,
        help='export reduction operations to a script file'
    )

    args = ap.parse_args(argv)

    if args.verbose:
        level = logging.DEBUG
    elif args.quiet:
        level = logging.WARNING
    elif args.silent:
        level = logging.ERROR
    else:
        level = logging.INFO

    logging.basicConfig(stream=sys.stdout, level=level)
    log = logging.getLogger('reduce.main')

    if not args.input.is_dir():
        m = proposal_run_input_pattern.match(str(args.input))

        if m:
            inp_data = open_run(proposal=m[1], run=m[2], data=m[3] or 'raw')
            log.info(f'Found proposal run at '
                     f'{Path(inp_data.files[0].filename).parent}')
    
    else:
        inp_data = RunDirectory(args.input)

    log.info(f'Opened data with {len(inp_data.all_sources)} sources with '
             f'{len(inp_data.train_ids)} trains across {len(inp_data.files)} '
             f'files')
    
    methods = {}

    for ep in iter_entry_points('exdf_tools.reduction_method'):
        method = ep.load()()
        method.compile(inp_data, args)

        if method:
            log.info(f'Applying {ep.name} with {len(method)} reduction '
                     f'operations')
            methods[ep.name] = method

    if args.to_script:
        with open(args.to_script, 'w') as f:
            for name, method in methods.items():
                f.write(f'# {name}\n')

                for instr in method._instructions:
                    f.write('{}: {}\n'.format(
                        instr[0], '; '.join([str(x) for x in instr[1:]])))

    seq = ReductionSequence(inp_data, methods)
    seq.link(args.output_scope)

    if args.output_scope == 'none':
        log.info(f'Not writing out any data files')
        return

    if args.output_layout == 'collection':
        args.output.mkdir(parents=True, exist_ok=True)

        log.debug(f'Writing collection to {args.output}')
        seq.write_collection(args.output, args.output_sequence_len)

    elif args.output_layout == 'voview':
        if args.output.is_file() or args.output.suffix == '.h5':
            output_path = args.output
        else:
            args.output.mkdir(parents=True, exist_ok=True)
            output_path = args.output / 'output.h5'

        log.debug(f'Writing voview to {output_path}')
        seq.write_voview(output_path)

    elif args.output_layout == 'collapsed':
        # Requires datasets to be in memory for now.
        log.warning('In the current implementation, the entire output data '
                    'must be loaded to memory')
        if input('Still continue? [y/n] ') != 'y':
            return

        args.output.mkdir(parents=True, exist_ok=True)

        log.debug(f'Writing collapsed collection to {args.output}')
        seq.write_collapsed(args.output, args.output_sequence_len)
        

if __name__ == '__main__':
    main()
