
from logging import getLogger

import numpy as np

from extra_data import by_id, SourceNameError
from exdf_tools.reduce import ReductionMethod


class ManualRemoval(ReductionMethod):
    @staticmethod
    def arguments(ap):
        group = ap.add_argument_group(
            'Manual removal',
            'Allows to manually remove sources, keys or trains.')

        group.add_argument(
            '--remove-sources',
            metavar='SRC', action='store', type=str, nargs='*',
            help='remove one or more sources'
        )

        group.add_argument(
            '--remove-keys',
            metavar='SRC,KEY', action='store', type=str, nargs='*',
            help='remove one or more keys from sources')

        group.add_argument(
            '--remove-trains',
            metavar='SLICE_EXPR', action='store', type=str, nargs='*',
            help='remove one or more trains')

    def compile(self, data, args):
        if args.remove_sources:
            # Easiest would be to copy file and just remove the source.
            for source_glob in args.remove_sources:
                self.remove_sources(source_glob)
    
        if args.remove_keys:
            # Easiest would be to copy file and just remove the key.
            for sourcekey_str in args.remove_keys:
                source_glob, key_glob = sourcekey_str.split(',')
                self.remove_keys(source_glob, key_glob)

        if args.remove_trains:
            for slice_expr in args.remove_trains:
                # TODO: Convert slice_expr top slice
                self.remove_trains('*', xd.by_id[args.remove_trains])


class PpuTrainSequences(ReductionMethod):  # p3160, r250
    log = getLogger('PpuTrainSequences')

    @staticmethod
    def arguments(ap):
        group = ap.add_argument_group(
            'PPU train sequences',
            'Allows to reduce data to those trains covered by PPU trigger '
            'sequences.')

        group.add_argument(
            '--ppu-select-trains-by',
            metavar='SRC', action='store', type=str,
            help='only select those trains covered by trigger sequences of '
                 'the given PPU device')

        group.add_argument(
            '--ppu-sliced-sources',
            metavar='SRC', action='store', type=str, nargs='*',
            default='HED_DET_AGIPD500K2G/DET/*CH0:xtdf',
            help='sources sliced by PPU train sequences, HED-AGIPD by default'
        )

        group.add_argument(
            '--ppu-train-offset',
            metavar='INT', action='store', type=int, default=0,
            help='offset between the train IDs in the PPU trigger sequences '
                 'and the actual trains picked, 0 by default')

    def compile(self, data, args):
        if not args.ppu_select_trains_by:
            return

        try:
            seq_start = data[args.ppu_select_trains_by,
                             'trainTrigger.sequenceStart.value'].ndarray()
        except (SourceNameError, KeyError):
            self.log.error(f'PPU source {args.ppu_select_trains_by} or a '
                           f'required key found in data')
            return

        # Only support single trains for now.
        train_ids = np.unique(seq_start)[1:] + args.ppu_train_offset

        self.log.info(f'Found {len(train_ids)} PPU sequences')

        for train_id in train_ids:
            self.select_trains(args.ppu_sliced_sources,
                               by_id[train_id:train_id+1])


class AgipdGain(ReductionMethod):
    log = getLogger('AgipdGain')

    @staticmethod
    def arguments(ap):
        group = ap.add_argument_group(
            'AGIPD gain',
            'Allows to compress, replace or remove AGIPD gain information.')

        group.add_argument(
            '--agipd-gain-remove-if-fixed',
            action='store_true',
            help='remove the chunks backing analog gain information from '
                 'AGIPD image data')

        group.add_argument(
            '--agipd-gain-sources', action='store', type=str,
            help='AGIPD sources for which to remove analog gain chunks, '
                 'detected using extra_data.components if left blank')

        group.add_argument(
            '--agipd-gain-control', action='store', type=str,
            help='AGIPD control source to obtain gain information from, '
                 'inferred from gain sources if left blank')

    @staticmethod
    def _find_agipd(data):
        from extra_data.components import AGIPD1M, AGIPD500K

        try:
            domain = AGIPD1M._find_detector_name(data)
        except SourceNameError:
            pass
        else:
            return domain

        try:
            domain = AGIPD500K._find_detector_name(data)
        except SourceNameError:
            pass
        else:
            return domain

    def compile(self, data, args):
        if not args.agipd_gain_remove_if_fixed:
            return

        if not args.agipd_gain_sources:
            domain = self._find_agipd(data)

            if domain is None:
                self.log.error('No AGIPD detected using '
                               'extra_data.components.AGIPD*, please pass '
                               'explicit gain sources')
                return

            self.log.info(f'Found {domain}')
            agipd_sources = f'{domain}/DET/*CH0:xtdf'
        
        else:
            agipd_sources = args.agipd_gain_sources

        try:
            if not args.agipd_gain_control:
                domain = agipd_sources[:agipd_sources.index('/')]

                if 'HED' in domain:
                    # Unfortunately HED uses a different domain for its
                    # AGIPD control devices, hardcode it here for
                    # convenience.
                    domain = domain.replace('_DET_', '_EXP_')

                control_source = f'{domain}/MDL/FPGA_COMP'
            else:
                control_source = args.agipd_gain_control

            gain_mode = data[control_source, 'gainModeIndex'].as_single_value()
        except (SourceNameError, KeyError):
            self.log.error(f'No control source {control_source} or required '
                           f'keys found in data')
            return

        #if gain_mode == 0:
        #    self.log.error('Data taken in adaptive mode, aborting')
        #    return

        self.partial_copy(agipd_sources, 'image.data',
                          np.s_[0, :, :], (64, 1, 512, 128))

